@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2017 The Qt Company Ltd.
+** Copyright (C) 2020 The Qt Company Ltd.
 ** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtQuick module of the Qt Toolkit.
@@ -41,6 +41,7 @@
 #include <QtCore/qmap.h>
 #include <QtGui/private/qguiapplication_p.h>
 #include <QtGui/private/qtouchdevice_p.h>
+#include <QtGui/private/qevent_p.h>
 #include <QtQuick/private/qquickitem_p.h>
 #include <QtQuick/private/qquickpointerhandler_p.h>
 #include <QtQuick/private/qquickwindow_p.h>
@@ -447,7 +448,7 @@ Item {
 */
 
 /*!
-    \qmlproperty int QtQuick::WheelEvent::inverted
+    \qmlproperty bool QtQuick::WheelEvent::inverted
 
     Returns whether the delta values delivered with the event are inverted.
 
@@ -864,8 +865,11 @@ void QQuickEventPoint::setGrabberItem(QQuickItem *grabber)
             QQuickWindowPrivate *windowPriv = QQuickWindowPrivate::get(grabber->window());
             windowPriv->sendUngrabEvent(oldGrabberItem, windowPriv->isDeliveringTouchAsMouse());
         }
-        for (QPointer<QQuickPointerHandler> passiveGrabber : m_passiveGrabbers)
-            passiveGrabber->onGrabChanged(passiveGrabber, OverrideGrabPassive, this);
+        if (grabber) {
+            for (QPointer<QQuickPointerHandler> passiveGrabber : m_passiveGrabbers)
+                if (passiveGrabber)
+                    passiveGrabber->onGrabChanged(passiveGrabber, OverrideGrabPassive, this);
+        }
     }
 }
 
@@ -910,10 +914,14 @@ void QQuickEventPoint::setGrabberPointerHandler(QQuickPointerHandler *grabber, b
                         passiveGrabber->onGrabChanged(grabber, OverrideGrabPassive, this);
                 }
             }
-            if (oldGrabberHandler)
+            if (oldGrabberHandler) {
                 oldGrabberHandler->onGrabChanged(oldGrabberHandler, (grabber ? CancelGrabExclusive : UngrabExclusive), this);
-            else if (oldGrabberItem && pointerEvent()->asPointerTouchEvent())
-                oldGrabberItem->touchUngrabEvent();
+            } else if (oldGrabberItem) {
+                if (pointerEvent()->asPointerTouchEvent())
+                    oldGrabberItem->touchUngrabEvent();
+                else if (pointerEvent()->asPointerMouseEvent())
+                    oldGrabberItem->mouseUngrabEvent();
+            }
             // touchUngrabEvent() can result in the grabber being set to null (MPTA does that, for example).
             // So set it again to ensure that final state is what we want.
             m_exclusiveGrabber = QPointer<QObject>(grabber);
@@ -1132,10 +1140,10 @@ void QQuickEventTouchPoint::reset(const QTouchEvent::TouchPoint &tp, ulong times
 struct PointVelocityData {
     QVector2D velocity;
     QPointF pos;
-    ulong timestamp;
+    ulong timestamp = 0;
 };
 
-typedef QMap<quint64, PointVelocityData*> PointDataForPointIdMap;
+typedef QMap<quint64, PointVelocityData> PointDataForPointIdMap;
 Q_GLOBAL_STATIC(PointDataForPointIdMap, g_previousPointData)
 static const int PointVelocityAgeLimit = 500; // milliseconds
 
@@ -1146,42 +1154,36 @@ static const int PointVelocityAgeLimit = 500; // milliseconds
 */
 QVector2D QQuickEventPoint::estimatedVelocity() const
 {
-    PointVelocityData *prevPoint = g_previousPointData->value(m_pointId);
-    if (!prevPoint) {
+    auto prevPointIt = g_previousPointData->find(m_pointId);
+    auto end = g_previousPointData->end();
+    if (prevPointIt == end) {
         // cleanup events older than PointVelocityAgeLimit
-        auto end = g_previousPointData->end();
         for (auto it = g_previousPointData->begin(); it != end; ) {
-            PointVelocityData *data = it.value();
-            if (m_timestamp - data->timestamp > PointVelocityAgeLimit) {
+            if (m_timestamp - it->timestamp > PointVelocityAgeLimit)
                 it = g_previousPointData->erase(it);
-                delete data;
-            } else {
+            else
                 ++it;
-            }
         }
-        // TODO optimize: stop this dynamic memory thrashing
-        prevPoint = new PointVelocityData;
-        prevPoint->velocity = QVector2D();
-        prevPoint->timestamp = 0;
-        prevPoint->pos = QPointF();
-        g_previousPointData->insert(m_pointId, prevPoint);
+        prevPointIt = g_previousPointData->insert(m_pointId, PointVelocityData());
     }
-    const ulong timeElapsed = m_timestamp - prevPoint->timestamp;
+
+    auto &prevPoint = prevPointIt.value();
+    const ulong timeElapsed = m_timestamp - prevPoint.timestamp;
     if (timeElapsed == 0)   // in case we call estimatedVelocity() twice on the same QQuickEventPoint
         return m_velocity;
 
     QVector2D newVelocity;
-    if (prevPoint->timestamp != 0)
-        newVelocity = QVector2D(m_scenePos - prevPoint->pos)/timeElapsed;
+    if (prevPoint.timestamp != 0)
+        newVelocity = QVector2D(m_scenePos - prevPoint.pos) / timeElapsed;
 
     // VERY simple kalman filter: does a weighted average
     // where the older velocities get less and less significant
     static const float KalmanGain = 0.7f;
     QVector2D filteredVelocity = newVelocity * KalmanGain + m_velocity * (1.0f - KalmanGain);
 
-    prevPoint->velocity = filteredVelocity;
-    prevPoint->pos = m_scenePos;
-    prevPoint->timestamp = m_timestamp;
+    prevPoint.velocity = filteredVelocity;
+    prevPoint.pos = m_scenePos;
+    prevPoint.timestamp = m_timestamp;
     return filteredVelocity;
 }
 
@@ -1429,6 +1431,135 @@ QQuickEventPoint *QQuickSinglePointEvent::point(int i) const
     return nullptr;
 }
 
+
+/*!
+    \qmltype PointerScrollEvent
+    \instantiates QQuickPointerScrollEvent
+    \inqmlmodule QtQuick
+    \ingroup qtquick-input-events
+    \brief Provides information about a scrolling event, such as from a mouse wheel.
+
+    \sa WheelHandler
+*/
+
+/*!
+    \internal
+    \class QQuickPointerScrollEvent
+*/
+
+/*!
+    \readonly
+    \qmlproperty PointerDevice QtQuick::PointerScrollEvent::device
+
+    This property holds the device that generated the event.
+*/
+
+/*!
+    \qmlproperty int QtQuick::PointerScrollEvent::buttons
+
+    This property holds the mouse buttons pressed when the wheel event was generated.
+
+    It contains a bitwise combination of:
+    \list
+    \li \l {Qt::LeftButton} {Qt.LeftButton}
+    \li \l {Qt::RightButton} {Qt.RightButton}
+    \li \l {Qt::MiddleButton} {Qt.MiddleButton}
+    \endlist
+*/
+
+/*!
+    \readonly
+    \qmlproperty int QtQuick::PointerScrollEvent::modifiers
+
+    This property holds the \l {Qt::KeyboardModifier}{keyboard modifier} keys
+    that were pressed immediately before the event occurred.
+
+    It contains a bitwise combination of the following flags:
+    \value Qt.NoModifier
+        No modifier key is pressed.
+    \value Qt.ShiftModifier
+        A Shift key on the keyboard is pressed.
+    \value Qt.ControlModifier
+        A Ctrl key on the keyboard is pressed.
+    \value Qt.AltModifier
+        An Alt key on the keyboard is pressed.
+    \value Qt.MetaModifier
+        A Meta key on the keyboard is pressed.
+    \value Qt.KeypadModifier
+        A keypad button is pressed.
+
+    For example, to react to a Shift key + Left mouse button click:
+    \qml
+    Item {
+        TapHandler {
+            onTapped: {
+                if ((event.button == Qt.LeftButton) && (event.modifiers & Qt.ShiftModifier))
+                    doSomething();
+            }
+        }
+    }
+    \endqml
+*/
+
+/*!
+    \qmlproperty point QtQuick::PointerScrollEvent::angleDelta
+
+    This property holds the distance that the wheel is rotated in wheel degrees.
+    The x and y cordinate of this property holds the delta in horizontal and
+    vertical orientation.
+
+    A positive value indicates that the wheel was rotated up/right;
+    a negative value indicates that the wheel was rotated down/left.
+
+    Most mouse types work in steps of 15 degrees, in which case the delta value is a
+    multiple of 120; i.e., 120 units * 1/8 = 15 degrees.
+*/
+
+/*!
+    \qmlproperty point QtQuick::PointerScrollEvent::pixelDelta
+
+    This property holds the delta in screen pixels and is available in platforms that
+    have high-resolution trackpads, such as \macos.
+    The x and y coordinates of this property hold the delta in horizontal and
+    vertical orientation. The value should be used directly to scroll content on screen.
+
+    For platforms without high-resolution touchpad support, pixelDelta will
+    always be (0,0), and angleDelta should be used instead.
+*/
+
+/*!
+    \qmlproperty bool QtQuick::PointerScrollEvent::hasAngleDelta
+
+    Returns whether the \l angleDelta property has a non-null value.
+*/
+
+/*!
+    \qmlproperty bool QtQuick::PointerScrollEvent::hasPixelDelta
+
+    Returns whether the \l pixelDelta property has a non-null value.
+*/
+
+/*!
+    \qmlproperty bool QtQuick::PointerScrollEvent::inverted
+
+    Returns whether the delta values delivered with the event are inverted.
+
+    Normally, a vertical wheel will produce a PointerScrollEvent with positive delta
+    values if the top of the wheel is rotating away from the hand operating it.
+    Similarly, a horizontal wheel movement will produce a PointerScrollEvent with
+    positive delta values if the top of the wheel is moved to the left.
+
+    However, on some platforms this is configurable, so that the same
+    operations described above will produce negative delta values (but with the
+    same magnitude). In a QML component (such as a tumbler or a slider) where
+    it is appropriate to synchronize the movement or rotation of an item with
+    the direction of the wheel, regardless of the system settings, the wheel
+    event handler can use the inverted property to decide whether to negate the
+    \l angleDelta or \l pixelDelta values.
+
+    \note Many platforms provide no such information. On such platforms,
+    \c inverted always returns false.
+*/
 QQuickPointerEvent *QQuickPointerScrollEvent::reset(QEvent *event)
 {
     m_event = static_cast<QInputEvent*>(event);
@@ -1447,7 +1578,7 @@ QQuickPointerEvent *QQuickPointerScrollEvent::reset(QEvent *event)
         m_synthSource = ev->source();
         m_inverted = ev->inverted();
 
-        m_point->reset(Qt::TouchPointMoved, ev->posF(), quint64(1) << 24, ev->timestamp()); // mouse has device ID 1
+        m_point->reset(Qt::TouchPointMoved, ev->position(), quint64(1) << 24, ev->timestamp()); // mouse has device ID 1
     }
 #endif
     // TODO else if (event->type() == QEvent::Scroll) ...
@@ -1827,6 +1958,7 @@ QTouchEvent *QQuickPointerTouchEvent::touchEventForItem(QQuickItem *item, bool i
     // but that would require changing tst_qquickwindow::touchEvent_velocity(): it expects transformed velocity
 
     bool anyPressOrReleaseInside = false;
+    bool anyStationaryWithModifiedPropertyInside = false;
     bool anyGrabber = false;
     QMatrix4x4 transformMatrix(QQuickItemPrivate::get(item)->windowToItemTransform());
     for (int i = 0; i < m_pointCount; ++i) {
@@ -1859,6 +1991,8 @@ QTouchEvent *QQuickPointerTouchEvent::touchEventForItem(QQuickItem *item, bool i
             anyPressOrReleaseInside = true;
         const QTouchEvent::TouchPoint *tp = touchPointById(p->pointId());
         if (tp) {
+            if (isInside && tp->d->stationaryWithModifiedProperty)
+                anyStationaryWithModifiedPropertyInside = true;
             eventStates |= tp->state();
             QTouchEvent::TouchPoint tpCopy = *tp;
             tpCopy.setPos(item->mapFromScene(tpCopy.scenePos()));
@@ -1872,7 +2006,8 @@ QTouchEvent *QQuickPointerTouchEvent::touchEventForItem(QQuickItem *item, bool i
 
     // Now touchPoints will have only points which are inside the item.
     // But if none of them were just pressed inside, and the item has no other reason to care, ignore them anyway.
-    if (eventStates == Qt::TouchPointStationary || touchPoints.isEmpty() || (!anyPressOrReleaseInside && !anyGrabber && !isFiltering))
+    if ((eventStates == Qt::TouchPointStationary && !anyStationaryWithModifiedPropertyInside) ||
+            touchPoints.isEmpty() || (!anyPressOrReleaseInside && !anyGrabber && !isFiltering))
         return nullptr;
 
     // if all points have the same state, set the event type accordingly
@@ -1936,6 +2071,10 @@ Q_QUICK_PRIVATE_EXPORT QDebug operator<<(QDebug dbg, const QQuickPointerEvent *e
 {
     QDebugStateSaver saver(dbg);
     dbg.nospace();
+    if (!event) {
+        dbg << "QQuickPointerEvent(0)";
+        return dbg;
+    }
     dbg << "QQuickPointerEvent(";
     dbg << event->timestamp();
     dbg << " dev:";
@@ -1956,6 +2095,10 @@ Q_QUICK_PRIVATE_EXPORT QDebug operator<<(QDebug dbg, const QQuickEventPoint *eve
 {
     QDebugStateSaver saver(dbg);
     dbg.nospace();
+    if (!event) {
+        dbg << "QQuickEventPoint(0)";
+        return dbg;
+    }
     dbg << "QQuickEventPoint(accepted:" << event->isAccepted()
         << " state:";
     QtDebugUtils::formatQEnum(dbg, event->state());
