@@ -43,6 +43,7 @@
 #include <QQmlIncubationController>
 #include <QTemporaryDir>
 #include <private/qqmlengine_p.h>
+#include <private/qqmltypedata_p.h>
 #include <QQmlAbstractUrlInterceptor>
 
 class tst_qqmlengine : public QQmlDataTest
@@ -82,6 +83,7 @@ private slots:
     void singletonInstance();
     void aggressiveGc();
     void cachedGetterLookup_qtbug_75335();
+    void createComponentOnSingletonDestruction();
 
 public slots:
     QObject *createAQObjectForOwnershipTest ()
@@ -427,7 +429,7 @@ void tst_qqmlengine::trimComponentCache()
     engine.setIncubationController(&componentCache);
 
     QQmlComponent component(&engine, testFileUrl(file));
-    QVERIFY(component.isReady());
+    QVERIFY2(component.isReady(), qPrintable(component.errorString()));
     QScopedPointer<QObject> object(component.create());
     QVERIFY(object != nullptr);
     QCOMPARE(object->property("success").toBool(), true);
@@ -741,13 +743,17 @@ public:
     CustomSelector(const QUrl &base):m_base(base){}
     virtual QUrl intercept(const QUrl &url, QQmlAbstractUrlInterceptor::DataType d)
     {
-        if (url.scheme() != QStringLiteral("file"))
+        if ((url.scheme() != QStringLiteral("file") && url.scheme() != QStringLiteral("qrc"))
+            || url.path().contains("QtQml"))
             return url;
         if (!m_interceptionPoints.contains(d))
             return url;
 
-        if (url.path().endsWith("Test.2/qmldir"))//Special case
-            return QUrl::fromLocalFile(m_base.path() + "interception/module/intercepted/qmldir");
+        if (url.path().endsWith("Test.2/qmldir")) {//Special case
+            QUrl url = m_base;
+            url.setPath(m_base.path() + "interception/module/intercepted/qmldir");
+            return url;
+        }
         // Special case: with 5.10 we always add the implicit import, so we need to explicitly handle this case now
         if (url.path().endsWith("intercepted/qmldir"))
             return url;
@@ -835,7 +841,7 @@ void tst_qqmlengine::urlInterceptor()
     QFETCH(QString, expectedAbsoluteUrl);
 
     QQmlEngine e;
-    e.setImportPathList(QStringList() << testFileUrl("interception/imports").toLocalFile());
+    e.addImportPath(testFileUrl("interception/imports").url());
     CustomSelector cs(testFileUrl(""));
     cs.m_interceptionPoints = interceptionPoint;
     e.setUrlInterceptor(&cs);
@@ -935,7 +941,7 @@ void tst_qqmlengine::cppSignalAndEval()
 {
     ObjectCaller objectCaller;
     QQmlEngine engine;
-    engine.rootContext()->setContextProperty(QLatin1Literal("CallerCpp"), &objectCaller);
+    engine.rootContext()->setContextProperty(QLatin1String("CallerCpp"), &objectCaller);
     QQmlComponent c(&engine);
     c.setData("import QtQuick 2.9\n"
               "Item {\n"
@@ -1014,6 +1020,65 @@ void tst_qqmlengine::singletonInstance()
         QCOMPARE(instance->metaObject()->className(), "JsSingleton");
     }
 
+    {
+        int data = 30;
+        auto id = qmlRegisterSingletonType<CppSingleton>("Qt.test",1,0,"CapturingLambda",[data](QQmlEngine*, QJSEngine*){ // register qobject singleton with capturing lambda
+                auto o = new CppSingleton;
+                o->setProperty("data", data);
+                return o;
+        });
+        QJSValue value = engine.singletonInstance<QJSValue>(id);
+        QVERIFY(!value.isUndefined());
+        QVERIFY(value.isQObject());
+        QObject *instance = value.toQObject();
+        QVERIFY(instance);
+        QCOMPARE(instance->metaObject()->className(), "CppSingleton");
+        QCOMPARE(instance->property("data"), data);
+    }
+    {
+        qmlRegisterSingletonType<CppSingleton>("Qt.test",1,0,"NotAmbiguous", [](QQmlEngine* qeng, QJSEngine* jeng) -> QObject* {return CppSingleton::create(qeng, jeng);}); // test that overloads for qmlRegisterSingleton are not ambiguous
+    }
+    {
+        // Register QObject* directly
+        CppSingleton single;
+        int id = qmlRegisterSingletonInstance("Qt.test", 1, 0, "CppOwned",
+                                                                &single);
+        QQmlEngine engine2;
+        CppSingleton *singlePtr = engine2.singletonInstance<CppSingleton *>(id);
+        QVERIFY(singlePtr);
+        QCOMPARE(&single, singlePtr);
+        QVERIFY(engine2.objectOwnership(singlePtr) == QQmlEngine::CppOwnership);
+    }
+
+    {
+        CppSingleton single;
+        QQmlEngine engineA;
+        QQmlEngine engineB;
+        int id = qmlRegisterSingletonInstance("Qt.test", 1, 0, "CppOwned", &single);
+        auto singlePtr = engineA.singletonInstance<CppSingleton *>(id);
+        QVERIFY(singlePtr);
+        singlePtr = engineA.singletonInstance<CppSingleton *>(id); // accessing the singleton multiple times from the same engine is fine
+        QVERIFY(singlePtr);
+        QTest::ignoreMessage(QtMsgType::QtCriticalMsg, "<Unknown File>: qmlRegisterSingletonType(): \"CppOwned\" is not available because the callback function returns a null pointer.");
+        QTest::ignoreMessage(QtMsgType::QtWarningMsg, "<Unknown File>: Singleton registered by registerSingletonInstance must only be accessed from one engine");
+        QCOMPARE(&single, singlePtr);
+        auto noSinglePtr = engineB.singletonInstance<CppSingleton *>(id);
+        QVERIFY(!noSinglePtr);
+    }
+
+    {
+        CppSingleton single;
+        QThread newThread {};
+        single.moveToThread(&newThread);
+        QCOMPARE(single.thread(), &newThread);
+        QQmlEngine engineB;
+        int id = qmlRegisterSingletonInstance("Qt.test", 1, 0, "CppOwned", &single);
+        QTest::ignoreMessage(QtMsgType::QtCriticalMsg, "<Unknown File>: qmlRegisterSingletonType(): \"CppOwned\" is not available because the callback function returns a null pointer.");
+        QTest::ignoreMessage(QtMsgType::QtWarningMsg, "<Unknown File>: Registered object must live in the same thread as the engine it was registered with");
+        auto noSinglePtr = engineB.singletonInstance<CppSingleton *>(id);
+        QVERIFY(!noSinglePtr);
+    }
+
     {
         // Invalid types
         QJSValue value;
@@ -1043,6 +1108,16 @@ void tst_qqmlengine::singletonInstance()
         SomeQObjectClass * instance = engine.singletonInstance<SomeQObjectClass*>(cppSingletonTypeId);
         QVERIFY(!instance);
     }
+
+    {
+        // deleted object
+        auto dayfly = new QObject{};
+        auto id = qmlRegisterSingletonInstance("Vanity", 1, 0, "Dayfly", dayfly);
+        delete dayfly;
+        QTest::ignoreMessage(QtMsgType::QtWarningMsg, "<Unknown File>: The registered singleton has already been deleted. Ensure that it outlives the engine.");
+        QObject *instance = engine.singletonInstance<QObject*>(id);
+        QVERIFY(!instance);
+    }
 }
 
 void tst_qqmlengine::aggressiveGc()
@@ -1067,6 +1142,36 @@ void tst_qqmlengine::cachedGetterLookup_qtbug_75335()
     QVERIFY(object != nullptr);
 }
 
+class EvilSingleton : public QObject
+{
+    Q_OBJECT
+public:
+    QPointer<QQmlEngine> m_engine;
+    EvilSingleton(QQmlEngine *engine) : m_engine(engine) {
+        connect(this, &QObject::destroyed, this, [this]() {
+            QQmlComponent component(m_engine);
+            component.setData("import QtQml 2.0\nQtObject {}", QUrl("file://Stuff.qml"));
+            QVERIFY(component.isReady());
+            QScopedPointer<QObject> obj(component.create());
+            QVERIFY(obj);
+        });
+    }
+};
+
+void tst_qqmlengine::createComponentOnSingletonDestruction()
+{
+    qmlRegisterSingletonType<EvilSingleton>("foo.foo", 1, 0, "Singleton",
+                                            [](QQmlEngine *engine, QJSEngine *) {
+        return new EvilSingleton(engine);
+    });
+
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("evilSingletonInstantiation.qml"));
+    QVERIFY(component.isReady());
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(obj);
+}
+
 QTEST_MAIN(tst_qqmlengine)
 
 #include "tst_qqmlengine.moc"
