@@ -28,6 +28,7 @@
 
 #include <QtTest/QtTest>
 #include <QtQml/qqmlengine.h>
+#include <QtQml/qqmlfile.h>
 #include <QtQml/qqmlnetworkaccessmanagerfactory.h>
 #include <QtQuick/qquickview.h>
 #include <QtQuick/qquickitem.h>
@@ -35,6 +36,7 @@
 #include <QtCore/qprocess.h>
 #endif
 #include <QtQml/private/qqmlengine_p.h>
+#include <QtQml/private/qqmltypedata_p.h>
 #include <QtQml/private/qqmltypeloader_p.h>
 #include "../../shared/testhttpserver.h"
 #include "../../shared/util.h"
@@ -55,6 +57,10 @@ private slots:
     void redirect();
     void qmlSingletonWithinModule();
     void multiSingletonModule();
+    void implicitComponentModule();
+    void qrcRootPathUrl();
+    void implicitImport();
+    void compositeSingletonCycle();
 };
 
 void tst_QQMLTypeLoader::testLoadComplete()
@@ -429,7 +435,7 @@ void tst_QQMLTypeLoader::redirect()
     component.loadUrl(server.urlString("/Load.qml"), QQmlComponent::Asynchronous);
     QTRY_VERIFY2(component.isReady(), qPrintable(component.errorString()));
 
-    QObject *object = component.create();
+    QScopedPointer<QObject> object {component.create()};
     QTRY_COMPARE(object->property("xy").toInt(), 323232);
 }
 
@@ -446,6 +452,31 @@ void tst_QQMLTypeLoader::qmlSingletonWithinModule()
     QVERIFY(obj->property("ok").toBool());
 }
 
+static void checkCleanCacheLoad(const QString &testCase)
+{
+#if QT_CONFIG(process)
+    const char *skipKey = "QT_TST_QQMLTYPELOADER_SKIP_MISMATCH";
+    if (qEnvironmentVariableIsSet(skipKey))
+        return;
+    for (int i = 0; i < 5; ++i) {
+        QProcess child;
+        child.setProgram(QCoreApplication::applicationFilePath());
+        child.setArguments(QStringList(testCase));
+        QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
+        env.insert(QLatin1String("QT_LOGGING_RULES"), QLatin1String("qt.qml.diskcache.debug=true"));
+        env.insert(QLatin1String(skipKey), QLatin1String("1"));
+        child.setProcessEnvironment(env);
+        child.start();
+        QVERIFY(child.waitForFinished());
+        QCOMPARE(child.exitCode(), 0);
+        QVERIFY(!child.readAllStandardOutput().contains("Checksum mismatch for cached version"));
+        QVERIFY(!child.readAllStandardError().contains("Checksum mismatch for cached version"));
+    }
+#else
+    Q_UNUSED(testCase);
+#endif
+}
+
 void tst_QQMLTypeLoader::multiSingletonModule()
 {
     qmlClearTypeRegistrations();
@@ -463,25 +494,52 @@ void tst_QQMLTypeLoader::multiSingletonModule()
     QVERIFY(!obj.isNull());
     QVERIFY(obj->property("ok").toBool());
 
-#if QT_CONFIG(process)
-    const char *skipKey = "QT_TST_QQMLTYPELOADER_SKIP_MISMATCH";
-    if (qEnvironmentVariableIsSet(skipKey))
-        return;
-    for (int i = 0; i < 5; ++i) {
-        QProcess child;
-        child.setProgram(QCoreApplication::applicationFilePath());
-        child.setArguments(QStringList(QLatin1String("multiSingletonModule")));
-        QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
-        env.insert(QLatin1String("QT_LOGGING_RULES"), QLatin1String("qt.qml.diskcache.debug=true"));
-        env.insert(QLatin1String(skipKey), QLatin1String("1"));
-        child.setProcessEnvironment(env);
-        child.start();
-        QVERIFY(child.waitForFinished());
-        QCOMPARE(child.exitCode(), 0);
-        QVERIFY(!child.readAllStandardOutput().contains("Checksum mismatch for cached version"));
-        QVERIFY(!child.readAllStandardError().contains("Checksum mismatch for cached version"));
-    }
-#endif
+    checkCleanCacheLoad(QLatin1String("multiSingletonModule"));
+}
+
+void tst_QQMLTypeLoader::implicitComponentModule()
+{
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("implicitcomponent.qml"));
+    QCOMPARE(component.status(), QQmlComponent::Ready);
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(!obj.isNull());
+
+    checkCleanCacheLoad(QLatin1String("implicitComponentModule"));
+}
+
+void tst_QQMLTypeLoader::qrcRootPathUrl()
+{
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("qrcRootPath.qml"));
+    QCOMPARE(component.status(), QQmlComponent::Ready);
+}
+
+void tst_QQMLTypeLoader::implicitImport()
+{
+    QQmlEngine engine;
+    engine.addImportPath(testFile("imports"));
+    QQmlComponent component(&engine, testFileUrl("implicitimporttest.qml"));
+    QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(!obj.isNull());
+}
+
+void tst_QQMLTypeLoader::compositeSingletonCycle()
+{
+    TestHTTPServer server;
+    QVERIFY2(server.listen(), qPrintable(server.errorString()));
+    QVERIFY(server.serveDirectory(dataDirectory()));
+
+    QQmlEngine engine;
+    QQmlComponent component(&engine);
+    engine.addImportPath(server.baseUrl().toString());
+    component.loadUrl(server.urlString("Com/Orga/Handlers/Handler.qml"), QQmlComponent::Asynchronous);
+    QTRY_VERIFY2(component.isReady(), qPrintable(component.errorString()));
+
+    QScopedPointer<QObject> object {component.create()};
+    QVERIFY(object);
+    QCOMPARE(qvariant_cast<QColor>(object->property("color")), QColorConstants::Black);
 }
 
 QTEST_MAIN(tst_QQMLTypeLoader)
